/* 
  Copyright (C) 2000 Pawel A. Gajda (mis@k2.net.pl)
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License published by
  the Free Software Foundation (see file COPYING for details).
*/

/*
  $Id: pkgset-merge.c,v 1.1.2.1 2003/04/02 14:54:08 mis Exp $
*/

#include <limits.h>
#include <stdint.h>
#include <string.h>
#include <errno.h>

#include <rpm/rpmlib.h>
#include <trurl/nassert.h>
#include <trurl/narray.h>
#include <trurl/nhash.h>

#include "i18n.h"
#include "log.h"
#include "pkg.h"
#include "pkgset.h"
#include "misc.h"

static
int find_replacement(struct pkgset *ps, struct pkg *pkg, struct pkg **rpkg)
{
    const struct capreq_idx_ent *ent;
    struct pkg *bypkg = NULL;


    *rpkg = NULL;
    
    if ((bypkg = pkgset_lookup_pkgn(ps, pkg->name)) &&
        pkg_cmp_name_evr(bypkg, pkg) > 0) {
        
        *rpkg = bypkg;
        
    } else if ((ent = capreq_idx_lookup(&ps->obs_idx, pkg->name))) {
        int i;
        
        for (i=0; i < ent->items; i++) {
            if (pkg_caps_obsoletes_pkg_caps(ent->pkgs[i], pkg) &&
                pkg_cmp_name_evr(ent->pkgs[i], pkg) > 0) {
                
                *rpkg = ent->pkgs[i];
                break;
            }
        }
    }
    
    if (*rpkg && strcmp((*rpkg)->name, pkg->name) != 0 &&
        (*rpkg)->caps && pkg->caps) {
        
        int i;
        for (i=0; i < n_array_size((*rpkg)->caps); i++) {
            struct capreq *cap = n_array_nth((*rpkg)->caps, i);

            if (strcmp(capreq_name(cap), (*rpkg)->name) == 0)
                continue;
            
            if (capreq_arr_find(pkg->caps, capreq_name(cap))) {
                msg(2, "%s is NOT obsoleted by %s (same %s)\n", pkg_snprintf_s(pkg),
                    pkg_snprintf_s0(*rpkg), capreq_name(cap));
                *rpkg = NULL;
                break;
            }
        }
    }

    return (*rpkg != NULL);
}

void pkgset_merge(struct pkgset *ps_dst, struct pkgset *ps_src,
                  unsigned psflags) /* TODO: psflags should be extractable
                                     from existing ps */
{
    struct pkg *pkg, *rpkg;
    tn_array   *rm_pkgs, *pkgs;
    struct pkgset *ps;
    struct pkgdir *pkgdir;
    int i;

    n_assert(n_array_size(ps_dst->pkgdirs) == 1);

    rm_pkgs = pkgs_array_new(1024);
    pkgs = pkgs_array_new(n_array_size(ps_dst->pkgdirs));

    for (i=0; i < n_array_size(ps_src->pkgs); i++) {
        pkg = n_array_nth(ps_src->pkgs, i);
        pkg_clr_badreqs(pkg);
        n_array_push(pkgs, pkg_link(pkg));
    }
    
    for (i=0; i < n_array_size(ps_dst->pkgs); i++) {
        pkg = n_array_nth(ps_dst->pkgs, i);
        pkg_clr_badreqs(pkg);
        
        if (!find_replacement(ps_src, pkg, &rpkg)) {
            n_array_push(pkgs, pkg_link(pkg));
            
        } else {
            n_array_push(rm_pkgs, pkg_link(pkg));
            n_array_push(rm_pkgs, pkg_link(rpkg));
            msg(2, "%s is obsoleted by %s\n", pkg_snprintf_s(pkg),
                pkg_snprintf_s0(rpkg));
        }
    }

    ps = pkgset_new(psflags | PSVERIFY_DEPS);
    ps->pkgs = n_ref(pkgs);
    pkgset_setup(ps, NULL);

    pkgdir = n_array_nth(ps_dst->pkgdirs, 0);
    for (i=0; i<n_array_size(ps->pkgs); i++) {
        pkg = n_array_nth(ps->pkgs, i);
        if (pkg_has_badreqs(pkg)) {
            if (pkg->pkgdir != pkgdir) {
                logn(LOGWARN, "%s: could not be moved\n", pkg_filename_s(pkg));
                
            } else {
                logn(LOGWARN, "%s: will be REMOVED from destination\n",
                     pkg_filename_s(pkg));
                printf("merge: rm %s\n", pkg_path_s(pkg));
            }
        } else if (pkg->pkgdir != pkgdir)
            printf("merge: mv %s %s\n", pkg_path_s(pkg), pkgdir->path);
    }
    
    while (n_array_size(rm_pkgs)) {
        pkg = n_array_shift(rm_pkgs);
        rpkg = n_array_shift(rm_pkgs);
        
        if (pkg_has_badreqs(rpkg)) {
            logn(LOGWARN, "%s: could not be moved\n", pkg_path_s(pkg));
            
        } else {
            printf("merge: rm %s\n", pkg_path_s(pkg));
        }
        

        pkg_free(pkg);
        pkg_free(rpkg);
    }
    
    n_array_free(pkgs);
    pkgset_free(ps);
}

